use crate::assert_or;
use crate::Error;
use axum::extract::Extension;
use axum::extract::Form;
use axum::response::Html;
use axum::response::Redirect;
use axum_sessions::extractors::ReadableSession;
use axum_sessions::extractors::WritableSession;
use pbkdf2::{
    password_hash::{rand_core::OsRng, PasswordHash, PasswordHasher, PasswordVerifier, SaltString},
    Pbkdf2,
};
use serde::Deserialize;
use sqlx::{Pool, Postgres};
use std::sync::Arc;
use tera::Context;
use tera::Tera;
use tracing::info;
use tracing::warn;

#[derive(Clone, Debug, Deserialize)]
pub struct RawUser {
    username: String,
    password: String,
}

pub async fn create_user(
    Form(data): Form<RawUser>,
    Extension(pool): Extension<Arc<Pool<Postgres>>>,
) -> Result<Redirect, Error> {
    assert_or!(
        data.username.is_empty(),
        "Username cannot be empty!".to_string()
    );
    assert_or!(
        data.password.is_empty(),
        "Password cannot be empty!".to_string()
    );

    let handle = tokio::task::spawn_blocking(move || {
        info!("Encrypting password");
        let salt = SaltString::generate(&mut OsRng);
        let password_hash = Pbkdf2.hash_password(data.password.as_bytes(), &salt);

        password_hash.map(|p| p.to_string())
    });

    let hash = handle.await??;

    info!("Inserting into database");
    sqlx::query("INSERT INTO users (username, password_hash) VALUES ($1, $2)")
        .bind(data.username)
        .bind(hash)
        .execute(&*pool)
        .await?;

    Ok(Redirect::to("/login"))
}

pub async fn login_backend(
    Form(data): Form<RawUser>,
    Extension(pool): Extension<Arc<Pool<Postgres>>>,
    mut session: WritableSession,
) -> Result<Redirect, Error> {
    assert_or!(
        data.username.is_empty(),
        "Username cannot be empty!".to_string()
    );
    info!("Fetching user hash");
    let hash: Option<(String,)> =
        sqlx::query_as("SELECT password_hash FROM users WHERE username=$1")
            .bind(&data.username)
            .fetch_optional(&*pool)
            .await?;

    let hash = hash.map(|(h,)| h).ok_or_else(|| {
        warn!("User was not found in the database");
        "User not found. Have you registered?".to_string()
    })?;

    let handle = tokio::task::spawn_blocking(move || {
        info!("Verifying password");
        let parsed_hash = PasswordHash::new(&hash)?;
        Pbkdf2
            .verify_password(data.password.as_bytes(), &parsed_hash)
            .map(|_| true)
    });

    handle.await??;

    // If we've gotten to this point, then the password was valid

    session.insert("logged_in_as", &data.username)?;

    Ok(Redirect::to("/"))
}

macro_rules! login_and_redirect {
    ($session:expr, $to:literal) => {
        if $session.get::<String>("logged_in_as").is_some() {
            ::tracing::warn!("User already logged in, redirecting to {}", $to);
            return Err(::axum::response::Redirect::to($to).into());
        }
    };
}

pub async fn register_form(
    Extension(tera): Extension<Arc<Tera>>,
    session: ReadableSession,
) -> Result<Html<String>, Error> {
    login_and_redirect!(session, "/");
    let rendered = tera.render("users/register.html", &Context::new())?;

    Ok(Html(rendered))
}

pub async fn login_form(
    Extension(tera): Extension<Arc<Tera>>,
    session: ReadableSession,
) -> Result<Html<String>, Error> {
    login_and_redirect!(session, "/");
    let rendered = tera.render("users/login.html", &Context::new())?;

    Ok(Html(rendered))
}
